# -*- coding: utf-8 -*-
# Copyright 2009-2012 Yelp
# Copyright 2013-2014 Yelp and Contributors
# Copyright 2015-2016 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions for compatibility with different version of hadoop."""
from distutils.version import LooseVersion
import logging
import os

from mrjob.py2 import string_types

# lists alternative names for jobconf variables
# full listing thanks to translation table in
# http://hadoop.apache.org/common/docs/current/hadoop-project-dist/hadoop-common/DeprecatedProperties.html # noqa

log = logging.getLogger(__name__)

_JOBCONF_DICT_LIST = [
    {'1.0': 'StorageId',
     '2.0': 'dfs.datanode.StorageId'},
    {'1.0': 'create.empty.dir.if.nonexist',
     '2.0': 'mapreduce.jobcontrol.createdir.ifnotexist'},
    {'1.0': 'dfs.access.time.precision',
     '2.0': 'dfs.namenode.accesstime.precision'},
    {'1.0': 'dfs.backup.address',
     '2.0': 'dfs.namenode.backup.address'},
    {'1.0': 'dfs.backup.http.address',
     '2.0': 'dfs.namenode.backup.http-address'},
    {'1.0': 'dfs.balance.bandwidthPerSec',
     '2.0': 'dfs.datanode.balance.bandwidthPerSec'},
    {'1.0': 'dfs.block.size',
     '2.0': 'dfs.blocksize'},
    {'1.0': 'dfs.client.buffer.dir',
     '2.0': 'fs.client.buffer.dir'},
    {'1.0': 'dfs.data.dir',
     '2.0': 'dfs.datanode.data.dir'},
    {'1.0': 'dfs.datanode.max.xcievers',
     '2.0': 'dfs.datanode.max.transfer.threads'},
    {'1.0': 'dfs.df.interval',
     '2.0': 'fs.df.interval'},
    {'1.0': 'dfs.http.address',
     '2.0': 'dfs.namenode.http-address'},
    {'1.0': 'dfs.https.address',
     '2.0': 'dfs.namenode.https-address'},
    {'1.0': 'dfs.https.client.keystore.resource',
     '2.0': 'dfs.client.https.keystore.resource'},
    {'1.0': 'dfs.https.need.client.auth',
     '2.0': 'dfs.client.https.need-auth'},
    {'1.0': 'dfs.max-repl-streams',
     '2.0': 'dfs.namenode.replication.max-streams'},
    {'1.0': 'dfs.max.objects',
     '2.0': 'dfs.namenode.max.objects'},
    {'1.0': 'dfs.name.dir',
     '2.0': 'dfs.namenode.name.dir'},
    {'1.0': 'dfs.name.dir.restore',
     '2.0': 'dfs.namenode.name.dir.restore'},
    {'1.0': 'dfs.name.edits.dir',
     '2.0': 'dfs.namenode.edits.dir'},
    {'1.0': 'dfs.permissions',
     '2.0': 'dfs.permissions.enabled'},
    {'1.0': 'dfs.permissions.supergroup',
     '2.0': 'dfs.permissions.superusergroup'},
    {'1.0': 'dfs.read.prefetch.size',
     '2.0': 'dfs.client.read.prefetch.size'},
    {'1.0': 'dfs.replication.considerLoad',
     '2.0': 'dfs.namenode.replication.considerLoad'},
    {'1.0': 'dfs.replication.interval',
     '2.0': 'dfs.namenode.replication.interval'},
    {'1.0': 'dfs.replication.min',
     '2.0': 'dfs.namenode.replication.min'},
    {'1.0': 'dfs.replication.pending.timeout.sec',
     '2.0': 'dfs.namenode.replication.pending.timeout-sec'},
    {'1.0': 'dfs.safemode.extension',
     '2.0': 'dfs.namenode.safemode.extension'},
    {'1.0': 'dfs.safemode.threshold.pct',
     '2.0': 'dfs.namenode.safemode.threshold-pct'},
    {'1.0': 'dfs.secondary.http.address',
     '2.0': 'dfs.namenode.secondary.http-address'},
    {'1.0': 'dfs.socket.timeout',
     '2.0': 'dfs.client.socket-timeout'},
    {'1.0': 'dfs.upgrade.permission',
     '2.0': 'dfs.namenode.upgrade.permission'},
    {'1.0': 'dfs.write.packet.size',
     '2.0': 'dfs.client-write-packet-size'},
    {'1.0': 'fs.checkpoint.dir',
     '2.0': 'dfs.namenode.checkpoint.dir'},
    {'1.0': 'fs.checkpoint.edits.dir',
     '2.0': 'dfs.namenode.checkpoint.edits.dir'},
    {'1.0': 'fs.checkpoint.period',
     '2.0': 'dfs.namenode.checkpoint.period'},
    {'1.0': 'fs.default.name',
     '2.0': 'fs.defaultFS'},
    {'1.0': 'hadoop.configured.node.mapping',
     '2.0': 'net.topology.configured.node.mapping'},
    {'1.0': 'hadoop.job.history.location',
     '2.0': 'mapreduce.jobtracker.jobhistory.location'},
    {'1.0': 'hadoop.native.lib',
     '2.0': 'io.native.lib.available'},
    {'1.0': 'hadoop.net.static.resolutions',
     '2.0': 'mapreduce.tasktracker.net.static.resolutions'},
    {'1.0': 'hadoop.pipes.command-file.keep',
     '2.0': 'mapreduce.pipes.commandfile.preserve'},
    {'1.0': 'hadoop.pipes.executable',
     '2.0': 'mapreduce.pipes.executable'},
    {'1.0': 'hadoop.pipes.executable.interpretor',
     '2.0': 'mapreduce.pipes.executable.interpretor'},
    {'1.0': 'hadoop.pipes.java.mapper',
     '2.0': 'mapreduce.pipes.isjavamapper'},
    {'1.0': 'hadoop.pipes.java.recordreader',
     '2.0': 'mapreduce.pipes.isjavarecordreader'},
    {'1.0': 'hadoop.pipes.java.recordwriter',
     '2.0': 'mapreduce.pipes.isjavarecordwriter'},
    {'1.0': 'hadoop.pipes.java.reducer',
     '2.0': 'mapreduce.pipes.isjavareducer'},
    {'1.0': 'hadoop.pipes.partitioner',
     '2.0': 'mapreduce.pipes.partitioner'},
    {'1.0': 'heartbeat.recheck.interval',
     '2.0': 'dfs.namenode.heartbeat.recheck-interval'},
    {'1.0': 'io.bytes.per.checksum',
     '2.0': 'dfs.bytes-per-checksum'},
    {'1.0': 'io.sort.factor',
     '2.0': 'mapreduce.task.io.sort.factor'},
    {'1.0': 'io.sort.mb',
     '2.0': 'mapreduce.task.io.sort.mb'},
    {'1.0': 'io.sort.spill.percent',
     '2.0': 'mapreduce.map.sort.spill.percent'},
    {'1.0': 'job.end.notification.url',
     '2.0': 'mapreduce.job.end-notification.url'},
    {'1.0': 'job.end.retry.attempts',
     '2.0': 'mapreduce.job.end-notification.retry.attempts'},
    {'1.0': 'job.end.retry.interval',
     '2.0': 'mapreduce.job.end-notification.retry.interval'},
    {'1.0': 'job.local.dir',
     '2.0': 'mapreduce.job.local.dir'},
    {'1.0': 'jobclient.completion.poll.interval',
     '2.0': 'mapreduce.client.completion.pollinterval'},
    {'1.0': 'jobclient.output.filter',
     '2.0': 'mapreduce.client.output.filter'},
    {'1.0': 'jobclient.progress.monitor.poll.interval',
     '2.0': 'mapreduce.client.progressmonitor.pollinterval'},
    {'1.0': 'keep.failed.task.files',
     '2.0': 'mapreduce.task.files.preserve.failedtasks'},
    {'1.0': 'keep.task.files.pattern',
     '2.0': 'mapreduce.task.files.preserve.filepattern'},
    {'1.0': 'key.value.separator.in.input.line',
     '2.0': 'mapreduce.input.keyvaluelinerecordreader.key.value.separator'},
    {'1.0': 'local.cache.size',
     '2.0': 'mapreduce.tasktracker.cache.local.size'},
    {'1.0': 'map.input.file',
     '2.0': 'mapreduce.map.input.file'},
    {'1.0': 'map.input.length',
     '2.0': 'mapreduce.map.input.length'},
    {'1.0': 'map.input.start',
     '2.0': 'mapreduce.map.input.start'},
    {'1.0': 'map.output.key.field.separator',
     '2.0': 'mapreduce.map.output.key.field.separator'},
    {'1.0': 'map.output.key.value.fields.spec',
     '2.0': 'mapreduce.fieldsel.map.output.key.value.fields.spec'},
    {'1.0': 'mapred.acls.enabled',
     '2.0': 'mapreduce.cluster.acls.enabled'},
    {'1.0': 'mapred.binary.partitioner.left.offset',
     '2.0': 'mapreduce.partition.binarypartitioner.left.offset'},
    {'1.0': 'mapred.binary.partitioner.right.offset',
     '2.0': 'mapreduce.partition.binarypartitioner.right.offset'},
    {'1.0': 'mapred.cache.archives',
     '2.0': 'mapreduce.job.cache.archives'},
    {'1.0': 'mapred.cache.archives.timestamps',
     '2.0': 'mapreduce.job.cache.archives.timestamps'},
    {'1.0': 'mapred.cache.files',
     '2.0': 'mapreduce.job.cache.files'},
    {'1.0': 'mapred.cache.files.timestamps',
     '2.0': 'mapreduce.job.cache.files.timestamps'},
    {'1.0': 'mapred.cache.localArchives',
     '2.0': 'mapreduce.job.cache.local.archives'},
    {'1.0': 'mapred.cache.localFiles',
     '2.0': 'mapreduce.job.cache.local.files'},
    {'1.0': 'mapred.child.tmp',
     '2.0': 'mapreduce.task.tmp.dir'},
    {'1.0': 'mapred.cluster.average.blacklist.threshold',
     '2.0': 'mapreduce.jobtracker.blacklist.average.threshold'},
    {'1.0': 'mapred.cluster.map.memory.mb',
     '2.0': 'mapreduce.cluster.mapmemory.mb'},
    {'1.0': 'mapred.cluster.max.map.memory.mb',
     '2.0': 'mapreduce.jobtracker.maxmapmemory.mb'},
    {'1.0': 'mapred.cluster.max.reduce.memory.mb',
     '2.0': 'mapreduce.jobtracker.maxreducememory.mb'},
    {'1.0': 'mapred.cluster.reduce.memory.mb',
     '2.0': 'mapreduce.cluster.reducememory.mb'},
    {'1.0': 'mapred.committer.job.setup.cleanup.needed',
     '2.0': 'mapreduce.job.committer.setup.cleanup.needed'},
    {'1.0': 'mapred.compress.map.output',
     '2.0': 'mapreduce.map.output.compress'},
    {'1.0': 'mapred.create.symlink',
     '2.0': 'mapreduce.job.cache.symlink.create'},
    {'1.0': 'mapred.data.field.separator',
     '2.0': 'mapreduce.fieldsel.data.field.separator'},
    {'1.0': 'mapred.debug.out.lines',
     '2.0': 'mapreduce.task.debugout.lines'},
    {'1.0': 'mapred.healthChecker.interval',
     '2.0': 'mapreduce.tasktracker.healthchecker.interval'},
    {'1.0': 'mapred.healthChecker.script.args',
     '2.0': 'mapreduce.tasktracker.healthchecker.script.args'},
    {'1.0': 'mapred.healthChecker.script.path',
     '2.0': 'mapreduce.tasktracker.healthchecker.script.path'},
    {'1.0': 'mapred.healthChecker.script.timeout',
     '2.0': 'mapreduce.tasktracker.healthchecker.script.timeout'},
    {'1.0': 'mapred.heartbeats.in.second',
     '2.0': 'mapreduce.jobtracker.heartbeats.in.second'},
    {'1.0': 'mapred.hosts',
     '2.0': 'mapreduce.jobtracker.hosts.filename'},
    {'1.0': 'mapred.hosts.exclude',
     '2.0': 'mapreduce.jobtracker.hosts.exclude.filename'},
    {'1.0': 'mapred.inmem.merge.threshold',
     '2.0': 'mapreduce.reduce.merge.inmem.threshold'},
    {'1.0': 'mapred.input.dir',
     '2.0': 'mapreduce.input.fileinputformat.inputdir'},
    {'1.0': 'mapred.input.dir.formats',
     '2.0': 'mapreduce.input.multipleinputs.dir.formats'},
    {'1.0': 'mapred.input.dir.mappers',
     '2.0': 'mapreduce.input.multipleinputs.dir.mappers'},
    {'1.0': 'mapred.input.pathFilter.class',
     '2.0': 'mapreduce.input.pathFilter.class'},
    {'1.0': 'mapred.jar',
     '2.0': 'mapreduce.job.jar'},
    {'1.0': 'mapred.job.classpath.archives',
     '2.0': 'mapreduce.job.classpath.archives'},
    {'1.0': 'mapred.job.classpath.files',
     '2.0': 'mapreduce.job.classpath.files'},
    {'1.0': 'mapred.job.id',
     '2.0': 'mapreduce.job.id'},
    {'1.0': 'mapred.job.map.memory.mb',
     '2.0': 'mapreduce.map.memory.mb'},
    {'1.0': 'mapred.job.name',
     '2.0': 'mapreduce.job.name'},
    {'1.0': 'mapred.job.priority',
     '2.0': 'mapreduce.job.priority'},
    {'1.0': 'mapred.job.queue.name',
     '2.0': 'mapreduce.job.queuename'},
    {'1.0': 'mapred.job.reduce.input.buffer.percent',
     '2.0': 'mapreduce.reduce.input.buffer.percent'},
    {'1.0': 'mapred.job.reduce.markreset.buffer.percent',
     '2.0': 'mapreduce.reduce.markreset.buffer.percent'},
    {'1.0': 'mapred.job.reduce.memory.mb',
     '2.0': 'mapreduce.reduce.memory.mb'},
    {'1.0': 'mapred.job.reduce.total.mem.bytes',
     '2.0': 'mapreduce.reduce.memory.totalbytes'},
    {'1.0': 'mapred.job.reuse.jvm.num.tasks',
     '2.0': 'mapreduce.job.jvm.numtasks'},
    {'1.0': 'mapred.job.shuffle.input.buffer.percent',
     '2.0': 'mapreduce.reduce.shuffle.input.buffer.percent'},
    {'1.0': 'mapred.job.shuffle.merge.percent',
     '2.0': 'mapreduce.reduce.shuffle.merge.percent'},
    {'1.0': 'mapred.job.tracker',
     '2.0': 'mapreduce.jobtracker.address'},
    {'1.0': 'mapred.job.tracker.handler.count',
     '2.0': 'mapreduce.jobtracker.handler.count'},
    {'1.0': 'mapred.job.tracker.history.completed.location',
     '2.0': 'mapreduce.jobtracker.jobhistory.completed.location'},
    {'1.0': 'mapred.job.tracker.http.address',
     '2.0': 'mapreduce.jobtracker.http.address'},
    {'1.0': 'mapred.job.tracker.jobhistory.lru.cache.size',
     '2.0': 'mapreduce.jobtracker.jobhistory.lru.cache.size'},
    {'1.0': 'mapred.job.tracker.persist.jobstatus.active',
     '2.0': 'mapreduce.jobtracker.persist.jobstatus.active'},
    {'1.0': 'mapred.job.tracker.persist.jobstatus.dir',
     '2.0': 'mapreduce.jobtracker.persist.jobstatus.dir'},
    {'1.0': 'mapred.job.tracker.persist.jobstatus.hours',
     '2.0': 'mapreduce.jobtracker.persist.jobstatus.hours'},
    {'1.0': 'mapred.job.tracker.retire.jobs',
     '2.0': 'mapreduce.jobtracker.retirejobs'},
    {'1.0': 'mapred.job.tracker.retiredjobs.cache.size',
     '2.0': 'mapreduce.jobtracker.retiredjobs.cache.size'},
    {'1.0': 'mapred.jobinit.threads',
     '2.0': 'mapreduce.jobtracker.jobinit.threads'},
    {'1.0': 'mapred.jobtracker.instrumentation',
     '2.0': 'mapreduce.jobtracker.instrumentation'},
    {'1.0': 'mapred.jobtracker.job.history.block.size',
     '2.0': 'mapreduce.jobtracker.jobhistory.block.size'},
    {'1.0': 'mapred.jobtracker.maxtasks.per.job',
     '2.0': 'mapreduce.jobtracker.maxtasks.perjob'},
    {'1.0': 'mapred.jobtracker.restart.recover',
     '2.0': 'mapreduce.jobtracker.restart.recover'},
    {'1.0': 'mapred.jobtracker.taskScheduler',
     '2.0': 'mapreduce.jobtracker.taskscheduler'},
    {'1.0': 'mapred.jobtracker.taskScheduler.maxRunningTasksPerJob',
     '2.0': 'mapreduce.jobtracker.taskscheduler.maxrunningtasks.perjob'},
    {'1.0': 'mapred.jobtracker.taskalloc.capacitypad',
     '2.0': 'mapreduce.jobtracker.taskscheduler.taskalloc.capacitypad'},
    {'1.0': 'mapred.join.expr',
     '2.0': 'mapreduce.join.expr'},
    {'1.0': 'mapred.join.keycomparator',
     '2.0': 'mapreduce.join.keycomparator'},
    {'1.0': 'mapred.lazy.output.format',
     '2.0': 'mapreduce.output.lazyoutputformat.outputformat'},
    {'1.0': 'mapred.line.input.format.linespermap',
     '2.0': 'mapreduce.input.lineinputformat.linespermap'},
    {'1.0': 'mapred.linerecordreader.maxlength',
     '2.0': 'mapreduce.input.linerecordreader.line.maxlength'},
    {'1.0': 'mapred.local.dir',
     '2.0': 'mapreduce.cluster.local.dir'},
    {'1.0': 'mapred.local.dir.minspacekill',
     '2.0': 'mapreduce.tasktracker.local.dir.minspacekill'},
    {'1.0': 'mapred.local.dir.minspacestart',
     '2.0': 'mapreduce.tasktracker.local.dir.minspacestart'},
    {'1.0': 'mapred.map.child.env',
     '2.0': 'mapreduce.map.env'},
    {'1.0': 'mapred.map.child.java.opts',
     '2.0': 'mapreduce.map.java.opts'},
    {'1.0': 'mapred.map.child.log.level',
     '2.0': 'mapreduce.map.log.level'},
    {'1.0': 'mapred.map.max.attempts',
     '2.0': 'mapreduce.map.maxattempts'},
    {'1.0': 'mapred.map.output.compression.codec',
     '2.0': 'mapreduce.map.output.compress.codec'},
    {'1.0': 'mapred.map.task.debug.script',
     '2.0': 'mapreduce.map.debug.script'},
    {'1.0': 'mapred.map.tasks',
     '2.0': 'mapreduce.job.maps'},
    {'1.0': 'mapred.map.tasks.speculative.execution',
     '2.0': 'mapreduce.map.speculative'},
    {'1.0': 'mapred.mapoutput.key.class',
     '2.0': 'mapreduce.map.output.key.class'},
    {'1.0': 'mapred.mapoutput.value.class',
     '2.0': 'mapreduce.map.output.value.class'},
    {'1.0': 'mapred.mapper.regex',
     '2.0': 'mapreduce.mapper.regex'},
    {'1.0': 'mapred.mapper.regex.group',
     '2.0': 'mapreduce.mapper.regexmapper..group'},
    {'1.0': 'mapred.max.map.failures.percent',
     '2.0': 'mapreduce.map.failures.maxpercent'},
    {'1.0': 'mapred.max.reduce.failures.percent',
     '2.0': 'mapreduce.reduce.failures.maxpercent'},
    {'1.0': 'mapred.max.split.size',
     '2.0': 'mapreduce.input.fileinputformat.split.maxsize'},
    {'1.0': 'mapred.max.tracker.blacklists',
     '2.0': 'mapreduce.jobtracker.tasktracker.maxblacklists'},
    {'1.0': 'mapred.max.tracker.failures',
     '2.0': 'mapreduce.job.maxtaskfailures.per.tracker'},
    {'1.0': 'mapred.merge.recordsBeforeProgress',
     '2.0': 'mapreduce.task.merge.progress.records'},
    {'1.0': 'mapred.min.split.size',
     '2.0': 'mapreduce.input.fileinputformat.split.minsize'},
    {'1.0': 'mapred.min.split.size.per.node',
     '2.0': 'mapreduce.input.fileinputformat.split.minsize.per.node'},
    {'1.0': 'mapred.min.split.size.per.rack',
     '2.0': 'mapreduce.input.fileinputformat.split.minsize.per.rack'},
    {'1.0': 'mapred.output.compress',
     '2.0': 'mapreduce.output.fileoutputformat.compress'},
    {'1.0': 'mapred.output.compression.codec',
     '2.0': 'mapreduce.output.fileoutputformat.compress.codec'},
    {'1.0': 'mapred.output.compression.type',
     '2.0': 'mapreduce.output.fileoutputformat.compress.type'},
    {'1.0': 'mapred.output.dir',
     '2.0': 'mapreduce.output.fileoutputformat.outputdir'},
    {'1.0': 'mapred.output.key.class',
     '2.0': 'mapreduce.job.output.key.class'},
    {'1.0': 'mapred.output.key.comparator.class',
     '2.0': 'mapreduce.job.output.key.comparator.class'},
    {'1.0': 'mapred.output.value.class',
     '2.0': 'mapreduce.job.output.value.class'},
    {'1.0': 'mapred.output.value.groupfn.class',
     '2.0': 'mapreduce.job.output.group.comparator.class'},
    {'1.0': 'mapred.permissions.supergroup',
     '2.0': 'mapreduce.cluster.permissions.supergroup'},
    {'1.0': 'mapred.pipes.user.inputformat',
     '2.0': 'mapreduce.pipes.inputformat'},
    {'1.0': 'mapred.reduce.child.env',
     '2.0': 'mapreduce.reduce.env'},
    {'1.0': 'mapred.reduce.child.java.opts',
     '2.0': 'mapreduce.reduce.java.opts'},
    {'1.0': 'mapred.reduce.child.log.level',
     '2.0': 'mapreduce.reduce.log.level'},
    {'1.0': 'mapred.reduce.max.attempts',
     '2.0': 'mapreduce.reduce.maxattempts'},
    {'1.0': 'mapred.reduce.parallel.copies',
     '2.0': 'mapreduce.reduce.shuffle.parallelcopies'},
    {'1.0': 'mapred.reduce.slowstart.completed.maps',
     '2.0': 'mapreduce.job.reduce.slowstart.completedmaps'},
    {'1.0': 'mapred.reduce.task.debug.script',
     '2.0': 'mapreduce.reduce.debug.script'},
    {'1.0': 'mapred.reduce.tasks',
     '2.0': 'mapreduce.job.reduces'},
    {'1.0': 'mapred.reduce.tasks.speculative.execution',
     '2.0': 'mapreduce.reduce.speculative'},
    {'1.0': 'mapred.seqbinary.output.key.class',
     '2.0': 'mapreduce.output.seqbinaryoutputformat.key.class'},
    {'1.0': 'mapred.seqbinary.output.value.class',
     '2.0': 'mapreduce.output.seqbinaryoutputformat.value.class'},
    {'1.0': 'mapred.shuffle.connect.timeout',
     '2.0': 'mapreduce.reduce.shuffle.connect.timeout'},
    {'1.0': 'mapred.shuffle.read.timeout',
     '2.0': 'mapreduce.reduce.shuffle.read.timeout'},
    {'1.0': 'mapred.skip.attempts.to.start.skipping',
     '2.0': 'mapreduce.task.skip.start.attempts'},
    {'1.0': 'mapred.skip.map.auto.incr.proc.count',
     '2.0': 'mapreduce.map.skip.proc-count.auto-incr'},
    {'1.0': 'mapred.skip.map.max.skip.records',
     '2.0': 'mapreduce.map.skip.maxrecords'},
    {'1.0': 'mapred.skip.on',
     '2.0': 'mapreduce.job.skiprecords'},
    {'1.0': 'mapred.skip.out.dir',
     '2.0': 'mapreduce.job.skip.outdir'},
    {'1.0': 'mapred.skip.reduce.auto.incr.proc.count',
     '2.0': 'mapreduce.reduce.skip.proc-count.auto-incr'},
    {'1.0': 'mapred.skip.reduce.max.skip.groups',
     '2.0': 'mapreduce.reduce.skip.maxgroups'},
    {'1.0': 'mapred.speculative.execution.slowNodeThreshold',
     '2.0': 'mapreduce.job.speculative.slownodethreshold'},
    {'1.0': 'mapred.speculative.execution.slowTaskThreshold',
     '2.0': 'mapreduce.job.speculative.slowtaskthreshold'},
    {'1.0': 'mapred.speculative.execution.speculativeCap',
     '2.0': 'mapreduce.job.speculative.speculativecap'},
    {'1.0': 'mapred.submit.replication',
     '2.0': 'mapreduce.client.submit.file.replication'},
    {'1.0': 'mapred.system.dir',
     '2.0': 'mapreduce.jobtracker.system.dir'},
    {'1.0': 'mapred.task.cache.levels',
     '2.0': 'mapreduce.jobtracker.taskcache.levels'},
    {'1.0': 'mapred.task.id',
     '2.0': 'mapreduce.task.attempt.id'},
    {'1.0': 'mapred.task.is.map',
     '2.0': 'mapreduce.task.ismap'},
    {'1.0': 'mapred.task.partition',
     '2.0': 'mapreduce.task.partition'},
    {'1.0': 'mapred.task.profile',
     '2.0': 'mapreduce.task.profile'},
    {'1.0': 'mapred.task.profile.maps',
     '2.0': 'mapreduce.task.profile.maps'},
    {'1.0': 'mapred.task.profile.params',
     '2.0': 'mapreduce.task.profile.params'},
    {'1.0': 'mapred.task.profile.reduces',
     '2.0': 'mapreduce.task.profile.reduces'},
    {'1.0': 'mapred.task.timeout',
     '2.0': 'mapreduce.task.timeout'},
    {'1.0': 'mapred.task.tracker.http.address',
     '2.0': 'mapreduce.tasktracker.http.address'},
    {'1.0': 'mapred.task.tracker.report.address',
     '2.0': 'mapreduce.tasktracker.report.address'},
    {'1.0': 'mapred.task.tracker.task-controller',
     '2.0': 'mapreduce.tasktracker.taskcontroller'},
    {'1.0': 'mapred.tasktracker.dns.interface',
     '2.0': 'mapreduce.tasktracker.dns.interface'},
    {'1.0': 'mapred.tasktracker.dns.nameserver',
     '2.0': 'mapreduce.tasktracker.dns.nameserver'},
    {'1.0': 'mapred.tasktracker.events.batchsize',
     '2.0': 'mapreduce.tasktracker.events.batchsize'},
    {'1.0': 'mapred.tasktracker.expiry.interval',
     '2.0': 'mapreduce.jobtracker.expire.trackers.interval'},
    {'1.0': 'mapred.tasktracker.indexcache.mb',
     '2.0': 'mapreduce.tasktracker.indexcache.mb'},
    {'1.0': 'mapred.tasktracker.instrumentation',
     '2.0': 'mapreduce.tasktracker.instrumentation'},
    {'1.0': 'mapred.tasktracker.map.tasks.maximum',
     '2.0': 'mapreduce.tasktracker.map.tasks.maximum'},
    {'1.0': 'mapred.tasktracker.memory_calculator_plugin',
     '2.0': 'mapreduce.tasktracker.resourcecalculatorplugin'},
    {'1.0': 'mapred.tasktracker.memorycalculatorplugin',
     '2.0': 'mapreduce.tasktracker.resourcecalculatorplugin'},
    {'1.0': 'mapred.tasktracker.reduce.tasks.maximum',
     '2.0': 'mapreduce.tasktracker.reduce.tasks.maximum'},
    {'1.0': 'mapred.tasktracker.taskmemorymanager.monitoring-interval',
     '2.0': 'mapreduce.tasktracker.taskmemorymanager.monitoringinterval'},
    {'1.0': 'mapred.tasktracker.tasks.sleeptime-before-sigkill',
     '2.0': 'mapreduce.tasktracker.tasks.sleeptimebeforesigkill'},
    {'1.0': 'mapred.temp.dir',
     '2.0': 'mapreduce.cluster.temp.dir'},
    {'1.0': 'mapred.text.key.comparator.options',
     '2.0': 'mapreduce.partition.keycomparator.options'},
    {'1.0': 'mapred.text.key.partitioner.options',
     '2.0': 'mapreduce.partition.keypartitioner.options'},
    {'1.0': 'mapred.textoutputformat.separator',
     '2.0': 'mapreduce.output.textoutputformat.separator'},
    {'1.0': 'mapred.tip.id',
     '2.0': 'mapreduce.task.id'},
    {'1.0': 'mapred.used.genericoptionsparser',
     '2.0': 'mapreduce.client.genericoptionsparser.used'},
    {'1.0': 'mapred.userlog.limit.kb',
     '2.0': 'mapreduce.task.userlog.limit.kb'},
    {'1.0': 'mapred.userlog.retain.hours',
     '2.0': 'mapreduce.job.userlog.retain.hours'},
    {'1.0': 'mapred.work.output.dir',
     '2.0': 'mapreduce.task.output.dir'},
    {'1.0': 'mapred.working.dir',
     '2.0': 'mapreduce.job.working.dir'},
    {'1.0': 'mapreduce.combine.class',
     '2.0': 'mapreduce.job.combine.class'},
    {'1.0': 'mapreduce.inputformat.class',
     '2.0': 'mapreduce.job.inputformat.class'},
    {'1.0': 'mapreduce.jobtracker.permissions.supergroup',
     '2.0': 'mapreduce.cluster.permissions.supergroup'},
    {'1.0': 'mapreduce.map.class',
     '2.0': 'mapreduce.job.map.class'},
    {'1.0': 'mapreduce.outputformat.class',
     '2.0': 'mapreduce.job.outputformat.class'},
    {'1.0': 'mapreduce.partitioner.class',
     '2.0': 'mapreduce.job.partitioner.class'},
    {'1.0': 'mapreduce.reduce.class',
     '2.0': 'mapreduce.job.reduce.class'},
    {'1.0': 'min.num.spills.for.combine',
     '2.0': 'mapreduce.map.combine.minspills'},
    {'1.0': 'reduce.output.key.value.fields.spec',
     '2.0': 'mapreduce.fieldsel.reduce.output.key.value.fields.spec'},
    {'1.0': 'security.job.submission.protocol.acl',
     '2.0': 'security.job.client.protocol.acl'},
    {'1.0': 'security.task.umbilical.protocol.acl',
     '2.0': 'security.job.task.protocol.acl'},
    {'1.0': 'sequencefile.filter.class',
     '2.0': 'mapreduce.input.sequencefileinputfilter.class'},
    {'1.0': 'sequencefile.filter.frequency',
     '2.0': 'mapreduce.input.sequencefileinputfilter.frequency'},
    {'1.0': 'sequencefile.filter.regex',
     '2.0': 'mapreduce.input.sequencefileinputfilter.regex'},
    {'1.0': 'session.id',
     '2.0': 'dfs.metrics.session-id'},
    {'1.0': 'slave.host.name',
     '2.0': 'dfs.datanode.hostname'},
    {'1.0': 'slave.host.name',
     '2.0': 'mapreduce.tasktracker.host.name'},
    {'1.0': 'tasktracker.contention.tracking',
     '2.0': 'mapreduce.tasktracker.contention.tracking'},
    {'1.0': 'tasktracker.http.threads',
     '2.0': 'mapreduce.tasktracker.http.threads'},
    {'1.0': 'topology.node.switch.mapping.impl',
     '2.0': 'net.topology.node.switch.mapping.impl'},
    {'1.0': 'topology.script.file.name',
     '2.0': 'net.topology.script.file.name'},
    {'1.0': 'topology.script.number.args',
     '2.0': 'net.topology.script.number.args'},
    {'1.0': 'user.name',
     '2.0': 'mapreduce.job.user.name'},
    {'1.0': 'webinterface.private.actions',
     '2.0': 'mapreduce.jobtracker.webinterface.trusted'},
]

# Handle compatibility for 0.x versions of Hadoop too
for jobconf_dict in _JOBCONF_DICT_LIST:
    jobconf_dict['0.20'] = jobconf_dict['1.0']
    jobconf_dict['0.21'] = jobconf_dict['2.0']


def _dict_list_to_compat_map(dict_list):
    # compat_map = {
    #   ...
    #   a: {'1.0': a, '2.0': b}
    #   ..
    # }
    compat_map = {}
    for version_dict in dict_list:
        for value in version_dict.values():
            compat_map[value] = version_dict
    return compat_map


_JOBCONF_MAP = _dict_list_to_compat_map(_JOBCONF_DICT_LIST)


def jobconf_from_env(variable, default=None):
    """Get the value of a jobconf variable from the runtime environment.

    For example, a :py:class:`~mrjob.job.MRJob` could use
    ``jobconf_from_env('map.input.file')`` to get the name of the file a
    mapper is reading input from.

    If the name of the jobconf variable is different in different versions of
    Hadoop (e.g. in Hadoop 2.0, ``map.input.file`` is
    ``mapreduce.map.input.file``), we'll automatically try all variants before
    giving up.

    Return *default* if that jobconf variable isn't set.
    """
    # try variable verbatim first
    name = variable.replace('.', '_')
    if name in os.environ:
        return os.environ[name]

    # try alternatives (arbitrary order)
    for var in _JOBCONF_MAP.get(variable, {}).values():
        name = var.replace('.', '_')
        if name in os.environ:
            return os.environ[name]

    return default


def jobconf_from_dict(jobconf, name, default=None):
    """Get the value of a jobconf variable from the given dictionary.

    :param dict jobconf: jobconf dictionary
    :param string name: name of the jobconf variable (e.g. ``'user.name'``)
    :param default: fallback value

    If the name of the jobconf variable is different in different versions of
    Hadoop (e.g. in Hadoop 2, ``map.input.file`` is
    ``mapreduce.map.input.file``), we'll automatically try all variants before
    giving up.

    Return *default* if that jobconf variable isn't set    """
    if name in jobconf:
        return jobconf[name]

    # try alternatives (arbitrary order)
    for alternative in _JOBCONF_MAP.get(name, {}).values():
        if alternative in jobconf:
            return jobconf[alternative]

    return default


def map_version(version, version_map):
    """Allows you to look up something by version (e.g. which jobconf variable
    to use, specifying only the versions where that value changed.

    *version* is a string

    *version_map* is a map from version (as a string) that a value changed
    to the new value.

    For efficiency, *version_map* can also be a list of tuples of
    ``(LooseVersion(version_as_string), value)``, with oldest versions first.

    If *version* is less than any version in *version_map*, use the value for
    the earliest version in *version_map*.
    """
    if version is None:
        raise TypeError

    if not version_map:
        raise ValueError

    if isinstance(version_map, dict):
        version_map = sorted((LooseVersion(k), v)
                             for k, v in version_map.items())

    req_version = LooseVersion(version)

    for min_version, value in reversed(version_map):
        if req_version >= min_version:
            return value
    else:
        return version_map[0][1]


def translate_jobconf(variable, version):
    """Translate *variable* to Hadoop version *version*. If it's not
    a variable we recognize, leave as-is.
    """
    if version is None:
        raise TypeError

    if variable in _JOBCONF_MAP:
        return map_version(version, _JOBCONF_MAP[variable])
    else:
        return variable


def translate_jobconf_for_all_versions(variable):
    """Get all known variants of the given jobconf variable.
    Unlike :py:func:`translate_jobconf`, returns a list."""
    return sorted(
        set([variable] + list(_JOBCONF_MAP.get(variable, {}).values())))


def translate_jobconf_dict(jobconf, hadoop_version=None):
    """Translates the configuration property name to match those that
    are accepted in hadoop_version. Prints a warning message if any
    configuration property name does not match the name in the hadoop
    version. Combines the original jobconf with the translated jobconf.

    :return: a map consisting of the original and translated configuration
             property names and values.
    """
    translated_jobconf = jobconf.copy()
    translation_warnings = {}

    for variable, value in jobconf.items():
        if hadoop_version:
            variants = [translate_jobconf(variable, hadoop_version)]
        else:
            variants = translate_jobconf_for_all_versions(variable)

        for variant in variants:
            if variant in jobconf:
                # this happens if variant == variable or
                # if the variant was in jobconf to start with
                continue

            translated_jobconf[variant] = value

            if hadoop_version:
                translation_warnings[variable] = variant

    if translation_warnings:
        log.warning("Detected hadoop configuration property names that"
                    " do not match hadoop version %s:"
                    "\nThe have been translated as follows\n %s",
                    hadoop_version,
                    '\n'.join([
                        "%s: %s" % (variable, variant) for variable, variant
                        in sorted(translation_warnings.items())]))

    return translated_jobconf


def uses_yarn(version):
    """Basically, is this Hadoop 2? This also handles versions in the
    zero series (0.23+) where YARN originated."""
    return (version_gte(version, '2') or
            version_gte(version, '0.23') and not version_gte(version, '1'))


def version_gte(version, cmp_version_str):
    """Return ``True`` if version >= *cmp_version_str*."""

    if not isinstance(version, string_types):
        raise TypeError('%r is not a string' % version)

    if not isinstance(cmp_version_str, string_types):
        raise TypeError('%r is not a string' % cmp_version_str)

    return LooseVersion(version) >= LooseVersion(cmp_version_str)



# Copyright 2009-2016 Yelp and Contributors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import fnmatch
import logging
import socket

try:
    import boto
    boto  # quiet "redefinition of unused ..." warning from pyflakes
except ImportError:
    # don't require boto; MRJobs don't actually need it when running
    # inside hadoop streaming
    boto = None

from mrjob.aws import s3_endpoint_for_region
from mrjob.fs.base import Filesystem
from mrjob.parse import is_s3_uri
from mrjob.parse import parse_s3_uri
from mrjob.parse import urlparse
from mrjob.retry import RetryWrapper
from mrjob.runner import GLOB_RE
from mrjob.util import read_file


log = logging.getLogger(__name__)

# if EMR throttles us, how long to wait (in seconds) before trying again?
_EMR_BACKOFF = 20
_EMR_BACKOFF_MULTIPLIER = 1.5
_EMR_MAX_TRIES = 20  # this takes about a day before we run out of tries


def s3_key_to_uri(s3_key):
    """Convert a boto Key object into an ``s3://`` URI"""
    return 's3://%s/%s' % (s3_key.bucket.name, s3_key.name)


def wrap_aws_conn(raw_conn):
    """Wrap a given boto Connection object so that it can retry when
    throttled."""
    def retry_if(ex):
        """Retry if we get a server error indicating throttling. Also
        handle spurious 505s that are thought to be part of a load
        balancer issue inside AWS."""
        return ((isinstance(ex, boto.exception.BotoServerError) and
                 ('Throttling' in ex.body or
                  'RequestExpired' in ex.body or
                  ex.status == 505)) or
                (isinstance(ex, socket.error) and
                 ex.args in ((104, 'Connection reset by peer'),
                             (110, 'Connection timed out'))))

    return RetryWrapper(raw_conn,
                        retry_if=retry_if,
                        backoff=_EMR_BACKOFF,
                        multiplier=_EMR_BACKOFF_MULTIPLIER,
                        max_tries=_EMR_MAX_TRIES)


class S3Filesystem(Filesystem):
    """Filesystem for Amazon S3 URIs. Typically you will get one of these via
    ``EMRJobRunner().fs``, composed with
    :py:class:`~mrjob.fs.ssh.SSHFilesystem` and
    :py:class:`~mrjob.fs.local.LocalFilesystem`.
    """

    def __init__(self, aws_access_key_id=None, aws_secret_access_key=None,
                 aws_security_token=None, s3_endpoint=None):
        """
        :param aws_access_key_id: Your AWS access key ID
        :param aws_secret_access_key: Your AWS secret access key
        :param aws_security_token: security token for use with temporary
                                   AWS credentials
        :param s3_endpoint: If set, always use this endpoint
        """
        super(S3Filesystem, self).__init__()
        self._s3_endpoint = s3_endpoint
        self._aws_access_key_id = aws_access_key_id
        self._aws_secret_access_key = aws_secret_access_key
        self._aws_security_token = aws_security_token

    def can_handle_path(self, path):
        return is_s3_uri(path)

    def du(self, path_glob):
        """Get the size of all files matching path_glob."""
        return sum(self.get_s3_key(uri).size for uri in self.ls(path_glob))

    def ls(self, path_glob):
        """Recursively list files on S3.

        *path_glob* can include ``?`` to match single characters or
        ``*`` to match 0 or more characters. Both ``?`` and ``*`` can match
        ``/``.

        .. versionchanged:: 0.5.0

            You no longer need a trailing slash to list "directories" on S3;
            both ``ls('s3://b/dir')`` and `ls('s3://b/dir/')` will list
            all keys starting with ``dir/``.
        """

        # clean up the  base uri to ensure we have an equal uri to boto (s3://)
        # just in case we get passed s3n://
        scheme = urlparse(path_glob).scheme

        # support globs
        glob_match = GLOB_RE.match(path_glob)

        # we're going to search for all keys starting with base_uri
        if glob_match:
            # cut it off at first wildcard
            base_uri = glob_match.group(1)
        else:
            base_uri = path_glob

        bucket_name, base_name = parse_s3_uri(base_uri)

        # allow subdirectories of the path/glob
        if path_glob and not path_glob.endswith('/'):
            dir_glob = path_glob + '/*'
        else:
            dir_glob = path_glob + '*'

        bucket = self.get_bucket(bucket_name)
        for key in bucket.list(base_name):
            uri = "%s://%s/%s" % (scheme, bucket_name, key.name)

            # enforce globbing
            if not (fnmatch.fnmatchcase(uri, path_glob) or
                    fnmatch.fnmatchcase(uri, dir_glob)):
                continue

            yield uri

    def md5sum(self, path):
        k = self.get_s3_key(path)
        return k.etag.strip('"')

    def _cat_file(self, filename):
        # stream lines from the s3 key
        s3_key = self.get_s3_key(filename)
        # yields_lines=False: warn read_file that s3_key yields chunks of bytes
        return read_file(
            s3_key_to_uri(s3_key), fileobj=s3_key, yields_lines=False)

    def mkdir(self, dest):
        """Make a directory. This does nothing on S3 because there are
        no directories.
        """
        pass

    def exists(self, path_glob):
        """Does the given path exist?

        If dest is a directory (ends with a "/"), we check if there are
        any files starting with that path.
        """
        # just fall back on ls(); it's smart
        try:
            paths = self.ls(path_glob)
        except boto.exception.S3ResponseError:
            paths = []
        return any(paths)

    def rm(self, path_glob):
        """Remove all files matching the given glob."""
        for uri in self.ls(path_glob):
            key = self.get_s3_key(uri)
            if key:
                log.debug('deleting ' + uri)
                key.delete()

    def touchz(self, dest):
        """Make an empty file in the given location. Raises an error if
        a non-empty file already exists in that location."""
        key = self.get_s3_key(dest)
        if key and key.size != 0:
            raise OSError('Non-empty file %r already exists!' % (dest,))

        self.make_s3_key(dest).set_contents_from_string('')

    # Utilities for interacting with S3 using S3 URIs.

    # Try to use the more general filesystem interface unless you really
    # need to do something S3-specific (e.g. setting file permissions)

    def make_s3_conn(self, region=''):
        """Create a connection to S3.

        :param region: region to use to choose S3 endpoint.

        If you are doing anything with buckets other than creating them
        or fetching basic metadata (name and location), it's best to use
        :py:meth:`get_bucket` because it chooses the appropriate S3 endpoint
        automatically.

        :return: a :py:class:`boto.s3.connection.S3Connection`, wrapped in a
                 :py:class:`mrjob.retry.RetryWrapper`
        """
        # give a non-cryptic error message if boto isn't installed
        if boto is None:
            raise ImportError('You must install boto to connect to S3')

        # self._s3_endpoint overrides region
        host = self._s3_endpoint or s3_endpoint_for_region(region)

        log.debug('creating S3 connection (to %s)' % host)

        raw_s3_conn = boto.connect_s3(
            aws_access_key_id=self._aws_access_key_id,
            aws_secret_access_key=self._aws_secret_access_key,
            host=host,
            security_token=self._aws_security_token)
        return wrap_aws_conn(raw_s3_conn)

    def get_bucket(self, bucket_name):
        """Get the bucket, connecting through the appropriate endpoint."""
        s3_conn = self.make_s3_conn()

        bucket = s3_conn.get_bucket(bucket_name)
        if self._s3_endpoint:
            return bucket

        try:
            location = bucket.get_location()
        except boto.exception.S3ResponseError as e:
            # it's possible to have access to a bucket but not access
            # to its location metadata. This happens on the 'elasticmapreduce'
            # bucket, for example (see #1170)
            if e.status == 403:
                log.warning('Could not infer endpoint for bucket %s; '
                            'assuming %s', bucket_name, s3_conn.host)
                return bucket

            raise

        if (s3_endpoint_for_region(location) != s3_conn.host):
            s3_conn = self.make_s3_conn(location)
            bucket = s3_conn.get_bucket(bucket_name)

        return bucket

    def get_s3_key(self, uri):
        """Get the boto Key object matching the given S3 uri, or
        return None if that key doesn't exist.

        uri is an S3 URI: ``s3://foo/bar``
        """
        bucket_name, key_name = parse_s3_uri(uri)

        try:
            bucket = self.get_bucket(bucket_name)
        except boto.exception.S3ResponseError as e:
            if e.status != 404:
                raise e
            key = None
        else:
            key = bucket.get_key(key_name)

        return key

    def make_s3_key(self, uri):
        """Create the given S3 key, and return the corresponding
        boto Key object.

        uri is an S3 URI: ``s3://foo/bar``
        """
        bucket_name, key_name = parse_s3_uri(uri)

        return self.get_bucket(bucket_name).new_key(key_name)

    def get_s3_keys(self, uri):
        """Get a stream of boto Key objects for each key inside
        the given dir on S3.

        uri is an S3 URI: ``s3://foo/bar``
        """
        bucket_name, key_prefix = parse_s3_uri(uri)
        bucket = self.get_bucket(bucket_name)
        for key in bucket.list(key_prefix):
            yield key

    def get_all_buckets(self):
        """Get a stream of all buckets owned by this user on S3."""
        return self.make_s3_conn().get_all_buckets()

    def create_bucket(self, bucket_name, location=''):
        """Create a bucket on S3, optionally setting location constraint."""
        return self.make_s3_conn().create_bucket(
            bucket_name, location=location)



# Copyright 2016 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from os.path import dirname
from os.path import join

from mrjob.job import MRJob
from mrjob.step import INPUT
from mrjob.step import OUTPUT
from mrjob.step import SparkScriptStep


class MRSparkScriptWordcount(MRJob):

    def steps(self):
        return [
            SparkScriptStep(
                script=join(dirname(__file__), 'spark_wordcount_script.py'),
                args=[INPUT, OUTPUT],
            ),
        ]


if __name__ == '__main__':
    MRSparkScriptWordcount.run()



# Copyright 2009-2012 Yelp and Contributors
# Copyright 2013 David Marin
# Copyright 2014-2016 Yelp and Contributors
# Copyright 2017 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import codecs
import logging
import os
import sys
import time
from io import BytesIO
from optparse import Option
from optparse import OptionError
from optparse import OptionGroup
from optparse import OptionParser

from mrjob.conf import combine_dicts
from mrjob.options import _add_basic_options
from mrjob.options import _add_job_options
from mrjob.options import _add_runner_options
from mrjob.options import _allowed_keys
from mrjob.options import _pick_runner_opts
from mrjob.options import _print_help_for_runner
from mrjob.options import _print_basic_help
from mrjob.step import StepFailedException
from mrjob.util import log_to_null
from mrjob.util import log_to_stream
from mrjob.util import parse_and_save_options


log = logging.getLogger(__name__)


# sentinel value; used when running MRJob as a script
_READ_ARGS_FROM_SYS_ARGV = '_READ_ARGS_FROM_SYS_ARGV'

# options only used to modify other options; don't pass to runners
_FAKE_OPTIONS = set([
    'no_emr_api_params',
])


class MRJobLauncher(object):
    """Handle running a MapReduce job on an executable from the command line.
    This class will eventually support running arbitrary executables; for now
    it only supports :py:class:`~mrjob.job.MRJob` subclasses. Up to v0.5 it is
    effectively part of the :py:class:`~mrjob.job.MRJob` class itself and
    should not be used externally in any way.
    """

    #: :py:class:`optparse.Option` subclass to use with the
    #: :py:class:`optparse.OptionParser` instance.
    OPTION_CLASS = Option

    def __init__(self, script_path=None, args=None, from_cl=False):
        """
        :param script_path: Path to script unless it's the first item of *args*
        :param args: Command line arguments
        :param from_cl: If not using sys.argv but still comming from the
                        command line (as opposed to a script, e.g. from
                        mrjob.cmd), don't override the option parser error
                        function (exit instead of throwing ValueError).
        """

        if script_path is not None:
            script_path = os.path.abspath(script_path)
        self._script_path = script_path

        # make sure we respect the $TZ (time zone) environment variable
        if hasattr(time, 'tzset'):
            time.tzset()

        self._passthrough_options = []
        self._file_options = []

        self.option_parser = OptionParser(usage=self._usage(),
                                          option_class=self.OPTION_CLASS,
                                          add_help_option=False)

        self.configure_options()

        # don't pass None to parse_args unless we're actually running
        # the MRJob script
        if args is _READ_ARGS_FROM_SYS_ARGV:
            self._cl_args = sys.argv[1:]
        else:
            # don't pass sys.argv to self.option_parser, and have it
            # raise an exception on error rather than printing to stderr
            # and exiting.
            self._cl_args = args or []

            def error(msg):
                raise ValueError(msg)

            if not from_cl:
                self.option_parser.error = error

        self.load_options(self._cl_args)

        # Make it possible to redirect stdin, stdout, and stderr, for testing
        # See sandbox(), below.
        #
        # These should always read/write bytes, not unicode. Generally,
        # on Python 2, sys.std* can read and write bytes, whereas on Python 3,
        # you need to use sys.std*.buffer (which doesn't exist on Python 2).
        #
        # However, certain Python 3 environments, such as Jupyter notebook,
        # act more like Python 2. See #1441.
        self.stdin = getattr(sys.stdin, 'buffer', sys.stdin)
        self.stdout = getattr(sys.stdout, 'buffer', sys.stdout)
        self.stderr = getattr(sys.stderr, 'buffer', sys.stderr)

    @classmethod
    def _usage(cls):
        """Command line usage string for this class"""
        return ("usage: mrjob run [script path|executable path|--help]"
                " [options] [input files]")

    def _print_help(self, options):
        """Print help for this job. This will either print runner
        or basic help. Override to allow other kinds of help."""
        if options.runner:
            _print_help_for_runner(options.runner, options.deprecated)
        else:
            _print_basic_help(self.option_parser,
                              self._usage(),
                              options.deprecated)

    @classmethod
    def run(cls, args=_READ_ARGS_FROM_SYS_ARGV):
        """Entry point for running job from the command-line.

        This is also the entry point when a mapper or reducer is run
        by Hadoop Streaming.

        Does one of:

        * Print step information (:option:`--steps`). See :py:meth:`show_steps`
        * Run a mapper (:option:`--mapper`). See :py:meth:`run_mapper`
        * Run a combiner (:option:`--combiner`). See :py:meth:`run_combiner`
        * Run a reducer (:option:`--reducer`). See :py:meth:`run_reducer`
        * Run the entire job. See :py:meth:`run_job`
        """
        # load options from the command line
        launcher = cls(args=args)
        launcher.run_job()

    def execute(self):
        # Launcher only runs jobs, doesn't do any Hadoop Streaming stuff
        self.run_job()

    def make_runner(self):
        """Make a runner based on command-line arguments, so we can
        launch this job on EMR, on Hadoop, or locally.

        :rtype: :py:class:`mrjob.runner.MRJobRunner`
        """
        if self.options.runner == 'emr':
            # avoid requiring dependencies (such as boto) for other runners
            from mrjob.emr import EMRJobRunner
            return EMRJobRunner(**self.emr_job_runner_kwargs())

        elif self.options.runner == 'dataproc':
            from mrjob.dataproc import DataprocJobRunner
            return DataprocJobRunner(**self.dataproc_job_runner_kwargs())

        elif self.options.runner == 'hadoop':
            from mrjob.hadoop import HadoopJobRunner
            return HadoopJobRunner(**self.hadoop_job_runner_kwargs())

        elif self.options.runner == 'inline':
            raise ValueError("inline is not supported in the multi-lingual"
                             " launcher.")

        else:
            # run locally by default
            from mrjob.local import LocalMRJobRunner
            return LocalMRJobRunner(**self.local_job_runner_kwargs())

    @classmethod
    def set_up_logging(cls, quiet=False, verbose=False, stream=None):
        """Set up logging when running from the command line. This is also
        used by the various command-line utilities.

        :param bool quiet: If true, don't log. Overrides *verbose*.
        :param bool verbose: If true, set log level to ``DEBUG`` (default is
                             ``INFO``)
        :param bool stream: Stream to log to (default is ``sys.stderr``)

        This will also set up a null log handler for boto, so we don't get
        warnings if boto tries to log about throttling and whatnot.
        """
        if quiet:
            log_to_null(name='mrjob')
            log_to_null(name='__main__')
        else:
            log_to_stream(name='mrjob', debug=verbose, stream=stream)
            log_to_stream(name='__main__', debug=verbose, stream=stream)

        log_to_null(name='boto')

    def run_job(self):
        """Run the all steps of the job, logging errors (and debugging output
        if :option:`--verbose` is specified) to STDERR and streaming the
        output to STDOUT.

        Called from :py:meth:`run`. You'd probably only want to call this
        directly from automated tests.
        """
        # self.stderr is strictly binary, need to wrap it so it's possible
        # to log to it in Python 3
        log_stream = codecs.getwriter('utf_8')(self.stderr)

        self.set_up_logging(quiet=self.options.quiet,
                            verbose=self.options.verbose,
                            stream=log_stream)

        with self.make_runner() as runner:
            try:
                runner.run()
            except StepFailedException as e:
                # no need for a runner stacktrace if step failed; runners will
                # log more useful information anyway
                log.error(str(e))
                sys.exit(1)

            if not self.options.no_output:
                for line in runner.stream_output():
                    self.stdout.write(line)
                self.stdout.flush()

    ### Command-line arguments ###

    def configure_options(self):
        """Define arguments for this script. Called from :py:meth:`__init__()`.

        Re-define to define custom command-line arguments or pass
        through existing ones::

            def configure_options(self):
                super(MRYourJob, self).configure_options

                self.add_passthrough_option(...)
                self.add_file_option(...)
                self.pass_through_option(...)
                ...
        """
        self.option_parser.add_option(
            '-h', '--help', dest='help', action='store_true', default=False,
            help='show this message and exit')

        self.option_parser.add_option(
            '--deprecated', dest='deprecated', action='store_true',
            default=False,
            help='include help for deprecated options')

        # protocol stuff
        self._proto_opt_group = OptionGroup(
            self.option_parser, 'Protocols')
        self.option_parser.add_option_group(self._proto_opt_group)

        _add_runner_options(
            self._proto_opt_group, set(['strict_protocols']))

        # options for running the job (any runner)
        self._runner_opt_group = OptionGroup(
            self.option_parser, 'Running the entire job')
        self.option_parser.add_option_group(self._runner_opt_group)

        _add_basic_options(self._runner_opt_group)
        _add_job_options(self._runner_opt_group)
        _add_runner_options(
            self._runner_opt_group,
            _pick_runner_opts('base') - set(['strict_protocols']))

        # options for inline/local runners
        self._local_opt_group = OptionGroup(
            self.option_parser,
            'Running locally (these apply when you set -r inline or -r local)')
        self.option_parser.add_option_group(self._local_opt_group)

        _add_runner_options(
            self._local_opt_group,
            _pick_runner_opts('local') - _pick_runner_opts('base'))

        # options common to Hadoop and EMR
        self._hadoop_emr_opt_group = OptionGroup(
            self.option_parser,
            'Running on Hadoop or EMR (these apply when you set -r hadoop or'
            ' -r emr)')
        self.option_parser.add_option_group(self._hadoop_emr_opt_group)

        _add_runner_options(
            self._hadoop_emr_opt_group,
            ((_pick_runner_opts('emr') & _pick_runner_opts('hadoop')) -
             _pick_runner_opts('base')))

        # options for running the job on Hadoop
        self._hadoop_opt_group = OptionGroup(
            self.option_parser,
            'Running on Hadoop (these apply when you set -r hadoop)')
        self.option_parser.add_option_group(self._hadoop_opt_group)

        _add_runner_options(
            self._hadoop_opt_group,
            (_pick_runner_opts('hadoop') -
             _pick_runner_opts('emr') - _pick_runner_opts('base')))

        # options for running the job on Dataproc or EMR
        self._dataproc_emr_opt_group = OptionGroup(
            self.option_parser,
            'Running on Dataproc or EMR (these apply when you set -r dataproc'
            ' or -r emr)')
        self.option_parser.add_option_group(self._dataproc_emr_opt_group)

        _add_runner_options(
            self._dataproc_emr_opt_group,
            ((_pick_runner_opts('dataproc') & _pick_runner_opts('emr')) -
             _pick_runner_opts('base')))

        # options for running the job on Dataproc
        self._dataproc_opt_group = OptionGroup(
            self.option_parser,
            'Running on Dataproc (these apply when you set -r dataproc)')
        self.option_parser.add_option_group(self._dataproc_opt_group)

        _add_runner_options(
            self._dataproc_opt_group,
            (_pick_runner_opts('dataproc') -
             _pick_runner_opts('emr') - _pick_runner_opts('base')))

        # options for running the job on EMR
        self._emr_opt_group = OptionGroup(
            self.option_parser,
            'Running on EMR (these apply when you set -r emr)')
        self.option_parser.add_option_group(self._emr_opt_group)

        _add_runner_options(
            self._emr_opt_group,
            (_pick_runner_opts('emr') - _pick_runner_opts('hadoop') -
             _pick_runner_opts('dataproc') - _pick_runner_opts('base')))

    def all_option_groups(self):
        log.warning('all_option_groups() is deprecated and will be removed'
                    ' in v0.6.0')

        return (self.option_parser, self._proto_opt_group,
                self._runner_opt_group, self._hadoop_opt_group,
                self._dataproc_emr_opt_group, self._hadoop_emr_opt_group,
                self._dataproc_opt_group, self._emr_opt_group,
                self._local_opt_group)

    def is_task(self):
        """True if this is a mapper, combiner, or reducer.

        This is mostly useful inside :py:meth:`load_options`, to disable
        loading options when we aren't running inside Hadoop Streaming.
        """
        return False

    def is_mapper_or_reducer(self):
        """The old name for :py:meth:`is_task`. Going away in v0.6.0"""
        log.warning('is_mapper_or_reducer() has been renamed to is_task().'
                    ' This alias will be removed in v0.6.0')
        return self.is_task()

    def add_passthrough_option(self, *args, **kwargs):
        """Function to create options which both the job runner
        and the job itself respect (we use this for protocols, for example).

        Use it like you would use :py:func:`optparse.OptionParser.add_option`::

            def configure_options(self):
                super(MRYourJob, self).configure_options()
                self.add_passthrough_option(
                    '--max-ngram-size', type='int', default=4, help='...')

        Specify an *opt_group* keyword argument to add the option to that
        :py:class:`OptionGroup` rather than the top-level
        :py:class:`OptionParser`.

        If you want to pass files through to the mapper/reducer, use
        :py:meth:`add_file_option` instead.

        If you want to pass through a built-in option (e.g. ``--runner``, use
        :py:meth:`pass_through_option` instead.
        """
        if 'opt_group' in kwargs:
            pass_opt = kwargs.pop('opt_group').add_option(*args, **kwargs)
        else:
            pass_opt = self.option_parser.add_option(*args, **kwargs)

        self._passthrough_options.append(pass_opt)

    def pass_through_option(self, opt_str):
        """Pass through a built-in option to tasks. For example, for
        tasks to see which runner launched them::

            def configure_options(self):
                super(MRYourJob, self).configure_options()
                self.pass_through_option('--runner')

            def mapper_init(self):
                if self.options.runner == 'emr':
                    ...

        *opt_str* can be a long option switch like ``--runner`` or a short
        one like ``-r``.

        .. versionadded:: 0.5.4
        """
        self._passthrough_options.append(
            self.option_parser.get_option(opt_str))

    def add_file_option(self, *args, **kwargs):
        """Add a command-line option that sends an external file
        (e.g. a SQLite DB) to Hadoop::

             def configure_options(self):
                super(MRYourJob, self).configure_options()
                self.add_file_option('--scoring-db', help=...)

        This does the right thing: the file will be uploaded to the working
        dir of the script on Hadoop, and the script will be passed the same
        option, but with the local name of the file in the script's working
        directory.

        We suggest against sending Berkeley DBs to your job, as
        Berkeley DB is not forwards-compatible (so a Berkeley DB that you
        construct on your computer may not be readable from within
        Hadoop). Use SQLite databases instead. If all you need is an on-disk
        hash table, try out the :py:mod:`sqlite3dbm` module.
        """
        pass_opt = self.option_parser.add_option(*args, **kwargs)

        if not pass_opt.type == 'string':
            raise OptionError(
                'passthrough file options must take strings' % pass_opt.type)

        if pass_opt.action not in ('store', 'append'):
            raise OptionError("passthrough file options must use the options"
                              " 'store' or 'append'")

        self._file_options.append(pass_opt)

    def _process_args(self, args):
        """mrjob.launch takes the first arg as the script path, but mrjob.job
        uses all args as input files. This method determines the behavior:
        MRJobLauncher takes off the first arg as the script path.
        """
        if not self._script_path:
            if len(args) < 1:
                self.option_parser.error('Must supply script path')
            else:
                self._script_path = os.path.abspath(args[0])
                self.args = args[1:]

    def load_options(self, args):
        """Load command-line options into ``self.options``,
        ``self._script_path``, and ``self.args``.

        Called from :py:meth:`__init__()` after :py:meth:`configure_options`.

        :type args: list of str
        :param args: a list of command line arguments. ``None`` will be
                     treated the same as ``[]``.

        Re-define if you want to post-process command-line arguments::

            def load_options(self, args):
                super(MRYourJob, self).load_options(args)

                self.stop_words = self.options.stop_words.split(',')
                ...
        """
        self.options, args = self.option_parser.parse_args(args)

        if self.options.help:
            self._print_help(self.options)
            sys.exit(0)

        self._process_args(args)

    def inline_job_runner_kwargs(self):
        """Keyword arguments to create create runners when
        :py:meth:`make_runner` is called, when we run a job locally
        (``-r inline``).

        :return: map from arg name to value

        Re-define this if you want finer control when running jobs locally.
        """
        return self._job_runner_kwargs_for_runner('inline')

    def local_job_runner_kwargs(self):
        """Keyword arguments to create create runners when
        :py:meth:`make_runner` is called, when we run a job locally
        (``-r local``).

        :return: map from arg name to value

        Re-define this if you want finer control when running jobs locally.
        """
        return self._job_runner_kwargs_for_runner('local')

    def emr_job_runner_kwargs(self):
        """Keyword arguments to create create runners when
        :py:meth:`make_runner` is called, when we run a job on EMR
        (``-r emr``).

        :return: map from arg name to value

        Re-define this if you want finer control when running jobs on EMR.
        """
        return self._job_runner_kwargs_for_runner('emr')

    def dataproc_job_runner_kwargs(self):
        """Keyword arguments to create create runners when
        :py:meth:`make_runner` is called, when we run a job on EMR
        (``-r emr``).

        :return: map from arg name to value

        Re-define this if you want finer control when running jobs on EMR.
        """
        return self._job_runner_kwargs_for_runner('dataproc')

    def hadoop_job_runner_kwargs(self):
        """Keyword arguments to create create runners when
        :py:meth:`make_runner` is called, when we run a job on EMR
        (``-r hadoop``).

        :return: map from arg name to value

        Re-define this if you want finer control when running jobs on hadoop.
        """
        return self._job_runner_kwargs_for_runner('hadoop')

    def _job_runner_kwargs_for_runner(self, runner_alias):
        """Helper method that powers the *_job_runner_kwargs()
        methods."""
        # user can no longer silently ignore switches by overriding
        # job_runner_kwargs()
        return combine_dicts(
            self._kwargs_from_switches(_allowed_keys(runner_alias)),
            self.job_runner_kwargs(),
        )

    def job_runner_kwargs(self):
        """Keyword arguments used to create runners when
        :py:meth:`make_runner` is called.

        :return: map from arg name to value

        Re-define this if you want finer control of runner initialization.

        You might find :py:meth:`mrjob.conf.combine_dicts` useful if you
        want to add or change lots of keyword arguments.
        """
        return combine_dicts(
            self._non_option_kwargs(),
            self._kwargs_from_switches(_allowed_keys('base')),
            self._job_kwargs(),
        )

    def _non_option_kwargs(self):
        """Keyword arguments to runner constructor that can't be set
        in mrjob.conf.

        These should match the (named) arguments to
        :py:meth:`~mrjob.runner.MRJobRunner.__init__`.
        """
        return dict(
            conf_paths=self.options.conf_paths,
            extra_args=self.generate_passthrough_arguments(),
            file_upload_args=self.generate_file_upload_args(),
            hadoop_input_format=self.hadoop_input_format(),
            hadoop_output_format=self.hadoop_output_format(),
            input_paths=self.args,
            mr_job_script=self._script_path,
            output_dir=self.options.output_dir,
            partitioner=self.partitioner(),
            stdin=self.stdin,
            step_output_dir=self.options.step_output_dir,
        )

    def _kwargs_from_switches(self, keys):
        return dict(
            (key, getattr(self.options, key))
            for key in keys if hasattr(self.options, key)
        )

    def _job_kwargs(self):
        """Keyword arguments to the runner class that can be specified
        by the job/launcher itself."""
        return dict(
            jobconf=self.jobconf(),
            libjars=self.libjars(),
            partitioner=self.partitioner(),
            sort_values=self.sort_values(),
        )

    ### Hooks for options defined by the job ###

    def hadoop_input_format(self):
        """See :py:meth:`mrjob.job.MRJob.hadoop_input_format`."""
        return None

    def hadoop_output_format(self):
        """See :py:meth:`mrjob.job.MRJob.hadoop_output_format`."""
        return None

    def jobconf(self):
        """See :py:meth:`mrjob.job.MRJob.jobconf`."""
        return {}

    def libjars(self):
        """See :py:meth:`mrjob.job.MRJob.libjars`."""
        return []

    def partitioner(self):
        """See :py:meth:`mrjob.job.MRJob.partitioner`."""
        return None

    def sort_values(self):
        """See :py:meth:`mrjob.job.MRJob.sort_values`."""
        return None

    ### More option stuff ###

    def generate_passthrough_arguments(self):
        """Returns a list of arguments to pass to subprocesses, either on
        hadoop or executed via subprocess.

        These are passed to :py:meth:`mrjob.runner.MRJobRunner.__init__`
        as *extra_args*.
        """
        arg_map = parse_and_save_options(self.option_parser, self._cl_args)
        output_args = []

        passthrough_dests = sorted(
            set(option.dest for option in self._passthrough_options))
        for option_dest in passthrough_dests:
            output_args.extend(arg_map.get(option_dest, []))

        return output_args

    def generate_file_upload_args(self):
        """Figure out file upload args to pass through to the job runner.

        Instead of generating a list of args, we're generating a list
        of tuples of ``('--argname', path)``

        These are passed to :py:meth:`mrjob.runner.MRJobRunner.__init__`
        as ``file_upload_args``.
        """
        file_upload_args = []

        master_option_dict = self.options.__dict__

        for opt in self._file_options:
            opt_prefix = opt.get_opt_string()
            opt_value = master_option_dict[opt.dest]

            if opt_value:
                paths = opt_value if opt.action == 'append' else [opt_value]
                for path in paths:
                    file_upload_args.append((opt_prefix, path))

        return file_upload_args

    ### Testing ###

    def sandbox(self, stdin=None, stdout=None, stderr=None):
        """Redirect stdin, stdout, and stderr for automated testing.

        You can set stdin, stdout, and stderr to file objects. By
        default, they'll be set to empty ``BytesIO`` objects.
        You can then access the job's file handles through ``self.stdin``,
        ``self.stdout``, and ``self.stderr``. See :ref:`testing` for more
        information about testing.

        You may call sandbox multiple times (this will essentially clear
        the file handles).

        ``stdin`` is empty by default. You can set it to anything that yields
        lines::

            mr_job.sandbox(stdin=BytesIO(b'some_data\\n'))

        or, equivalently::

            mr_job.sandbox(stdin=[b'some_data\\n'])

        For convenience, this sandbox() returns self, so you can do::

            mr_job = MRJobClassToTest().sandbox()

        Simple testing example::

            mr_job = MRYourJob.sandbox()
            self.assertEqual(list(mr_job.reducer('foo', ['a', 'b'])), [...])

        More complex testing example::

            from BytesIO import BytesIO

            from mrjob.parse import parse_mr_job_stderr
            from mrjob.protocol import JSONProtocol

            mr_job = MRYourJob(args=[...])

            fake_input = '"foo"\\t"bar"\\n"foo"\\t"baz"\\n'
            mr_job.sandbox(stdin=BytesIO(fake_input))

            mr_job.run_reducer(link_num=0)

            self.assertEqual(mrjob.stdout.getvalue(), ...)
            self.assertEqual(parse_mr_job_stderr(mr_job.stderr), ...)
        """
        self.stdin = stdin or BytesIO()
        self.stdout = stdout or BytesIO()
        self.stderr = stderr or BytesIO()

        return self

    def __getattr__(self, name):
        if name.startswith('_') and name.endswith('_opt_group'):
            log.warning('The %s attribute is deprecated and will be removed'
                        ' in v0.6.0' % name)
            return getattr(self, name[1:])
        else:
            raise AttributeError(name)


def _dests(opt_group):
    return set(s.dest for s in opt_group.option_list)


if __name__ == '__main__':
    MRJobLauncher.run()



# -*- coding: utf-8 -*-
# Copyright 2015-2016 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for ls()ing and cat()ing logs without raising exceptions."""
from logging import getLogger

from mrjob.py2 import to_string

from .ids import _sort_by_recency

log = getLogger(__name__)


def _cat_log(fs, path):
    """fs.cat() the given log, converting lines to strings, and logging
    errors."""
    try:
        if not fs.exists(path):
            return
        for line in fs.cat(path):
            yield to_string(line)
    except (IOError, OSError) as e:
        log.warning("couldn't cat() %s: %r" % (path, e))


def _ls_logs(fs, log_dir_stream, matcher, **kwargs):
    """Return a list matches against log files. Used to implement
    ``_ls_*_logs()`` functions.

    This yields dictionaries with ``path`` set to matching log path, and
    other information (e.g. corresponding job_id) returned by *matcher*

    *fs* is a :py:class:`mrjob.fs.Filesystem`

    *log_dir_stream* is a sequence of lists of log dirs. The idea is that
    there may be copies of the same logs in multiple places (e.g.
    on S3 and by SSHing into nodes) and we want to list them all without
    finding duplicate copies. This function will go through the lists of
    log dirs in turn, stopping if it finds any matches from a list.

    *matcher* is a function that takes (log_path, **kwargs)
    and returns either None (no match) or a dictionary with information
    about the path (e.g. the corresponding job_id). It's okay to return
    an empty dict.
    """
    # wrapper for fs.ls() that turns IOErrors into warnings
    def _fs_ls(path):
        try:
            if fs.exists(log_dir):
                for path in fs.ls(log_dir):
                    yield path
        except (IOError, OSError) as e:
            log.warning("couldn't ls() %s: %r" % (log_dir, e))

    for log_dirs in log_dir_stream:
        if isinstance(log_dirs, str):
            raise TypeError

        matches = []

        for log_dir in log_dirs:
            for path in _fs_ls(log_dir):
                m = matcher(path, **kwargs)
                if m is not None:
                    m['path'] = path
                    matches.append(m)

        if matches:
            return _sort_by_recency(matches)

    return []


def _logs_exist(fs, path):
    """Do ``fs.exists(path)``, and return ``None`` if it raises ``IOError``"""
    try:
        return fs.exists(path)
    except IOError:
        return None



# -*- coding: utf-8 -*-
# Copyright 2013 Lyft
# Copyright 2015-2016 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""General information about Amazon Web Services, such as region-to-endpoint
mappings.
"""

### EC2 Instances ###

# map from instance type to number of compute units, from:
# - https://aws.amazon.com/ec2/pricing/
# - http://aws.amazon.com/ec2/previous-generation/
EC2_INSTANCE_TYPE_TO_COMPUTE_UNITS = {
    'c1.medium': 5,
    'c1.xlarge': 20,
    'c3.2xlarge': 28,
    'c3.4xlarge': 55,
    'c3.8xlarge': 108,
    'c3.large': 7,
    'c3.xlarge': 14,
    'c4.2xlarge': 31,
    'c4.4xlarge': 62,
    'c4.8xlarge': 132,
    'c4.large': 8,
    'c4.xlarge': 16,
    'cc1.4xlarge': 33.5,
    'cc2.8xlarge': 88,
    'cg1.4xlarge': 33.5,
    'cr1.8xlarge': 88,
    'd2.2xlarge': 27,
    'd2.4xlarge': 56,
    'd2.8xlarge': 116,
    'd2.xlarge': 14,
    'g2.2xlarge': 26,
    'g2.8xlarge': 104,
    'hi1.4xlarge': 35,
    'hs1.8xlarge': 35,
    'i2.2xlarge': 27,
    'i2.4xlarge': 53,
    'i2.8xlarge': 104,
    'i2.xlarge': 14,
    'm1.large': 4,
    'm1.medium': 2,
    'm1.small': 1,
    'm1.xlarge': 8,
    'm2.2xlarge': 13,
    'm2.4xlarge': 26,
    'm2.xlarge': 6.5,
    'm3.2xlarge': 26,
    'm3.large': 6.5,
    'm3.medium': 3,
    'm3.xlarge': 13,
    'm4.10xlarge': 124.5,
    'm4.2xlarge': 26,
    'm4.4xlarge': 53.5,
    'm4.large': 6.5,
    'm4.xlarge': 13,
    'r3.2xlarge': 26,
    'r3.4xlarge': 52,
    'r3.8xlarge': 104,
    'r3.large': 6.5,
    'r3.xlarge': 13,
    't1.micro': 1,
    # t2 units are "burstable", and receive a certain number of "credits"
    # (CPU-minutes) per hour. MapReduce usage is pretty much continuous, so
    # just rating them by how much they can use in one hour.
    't2.large': 0.6,
    't2.medium': 0.4,
    't2.micro': 0.1,
    't2.nano': 0.05,
    't2.small': 0.2,
    'x1.32xlarge': 349,
}

# map from instance type to GB of memory
# from http://aws.amazon.com/ec2/instance-types/
# and http://aws.amazon.com/ec2/previous-generation/
EC2_INSTANCE_TYPE_TO_MEMORY = {
    'c1.medium': 1.7,
    'c1.xlarge': 7,
    'c3.2xlarge': 15,
    'c3.4xlarge': 30,
    'c3.8xlarge': 60,
    'c3.large': 3.75,
    'c3.xlarge': 7.5,
    'c4.2xlarge': 15,
    'c4.4xlarge': 30,
    'c4.8xlarge': 60,
    'c4.large': 3.75,
    'c4.xlarge': 7.5,
    'cc1.4xlarge': 23,
    'cc2.8xlarge': 60.5,
    'cg1.4xlarge': 22.5,
    'cr1.8xlarge': 244,
    'd2.2xlarge': 61,
    'd2.4xlarge': 122,
    'd2.8xlarge': 244,
    'd2.xlarge': 30.5,
    'g2.2xlarge': 15,
    'g2.8xlarge': 60,
    'hi1.4xlarge': 60.5,
    'hs1.8xlarge': 117,
    'i2.2xlarge': 61,
    'i2.4xlarge': 122,
    'i2.8xlarge': 244,
    'i2.xlarge': 30.5,
    'm1.large': 7.5,
    'm1.medium': 3.75,
    'm1.small': 1.7,
    'm1.xlarge': 17.1,
    'm2.2xlarge': 34.2,
    'm2.4xlarge': 68.4,
    'm2.xlarge': 17.5,
    'm3.2xlarge': 30,
    'm3.large': 7.5,
    'm3.medium': 3.75,
    'm3.xlarge': 15,
    'm4.10xlarge': 160,
    'm4.2xlarge': 32,
    'm4.4xlarge': 64,
    'm4.large': 8,
    'm4.xlarge': 16,
    'r3.2xlarge': 61,
    'r3.4xlarge': 122,
    'r3.8xlarge': 244,
    'r3.large': 15,
    'r3.xlarge': 30.5,
    't1.micro': 0.615,
    't2.large': 8,
    't2.medium': 4,
    't2.micro': 1,
    't2.nano': 0.5,
    't2.small': 2,
    'x1.32xlarge': 1952,
}


### Regions ###

# Based on http://docs.aws.amazon.com/general/latest/gr/rande.html

# See Issue #658 for why we don't just let boto handle this.


# where to connect to EMR. The docs say
# elasticmapreduce.<region>.amazonaws.com, but the SSL certificates,
# they tell a different story. See Issue #621.

# where the AWS docs say to connect to EMR
_EMR_REGION_ENDPOINT = 'elasticmapreduce.%(region)s.amazonaws.com'
# the host that currently works with EMR's SSL certificate
_EMR_REGION_SSL_HOST = '%(region)s.elasticmapreduce.amazonaws.com'
# the regionless endpoint doesn't have SSL issues
_EMR_REGIONLESS_ENDPOINT = 'elasticmapreduce.amazonaws.com'

# where to connect to S3
_S3_REGION_ENDPOINT = 's3-%(region)s.amazonaws.com'
_S3_REGIONLESS_ENDPOINT = 's3.amazonaws.com'

# us-east-1 doesn't have its own endpoint or need bucket location constraints
_S3_REGION_WITH_NO_LOCATION_CONSTRAINT = 'us-east-1'


# "EU" is an alias for the eu-west-1 region
_ALIAS_TO_REGION = {
    'eu': 'eu-west-1',
}

# The region to assume if none is specified
_DEFAULT_REGION = 'us-east-1'


def _fix_region(region):
    """Convert "EU" to "eu-west-1", None to '', and convert to lowercase."""
    region = (region or '').lower()
    return _ALIAS_TO_REGION.get(region) or region


def emr_endpoint_for_region(region):
    """Get the host for Elastic MapReduce in the given AWS region."""
    region = _fix_region(region)

    if not region:
        return _EMR_REGIONLESS_ENDPOINT
    else:
        return _EMR_REGION_ENDPOINT % {'region': region}


def emr_ssl_host_for_region(region):
    """Get the host for Elastic MapReduce that matches their SSL cert
    for the given region. (See Issue #621.)"""
    region = _fix_region(region)

    if not region:
        return _EMR_REGIONLESS_ENDPOINT
    else:
        return _EMR_REGION_SSL_HOST % {'region': region}


def s3_endpoint_for_region(region):
    """Get the host for S3 in the given AWS region.

    This will accept ``''`` for region as well, so it's fine to
    use location constraint in place of region.
    """
    region = _fix_region(region)

    if not region or region == _S3_REGION_WITH_NO_LOCATION_CONSTRAINT:
        return _S3_REGIONLESS_ENDPOINT
    else:
        return _S3_REGION_ENDPOINT % {'region': region}


def s3_location_constraint_for_region(region):
    """Get the location constraint an S3 bucket needs so that other AWS
    services can connect to it in the given region."""
    region = _fix_region(region)

    if not region or region == _S3_REGION_WITH_NO_LOCATION_CONSTRAINT:
        return ''
    else:
        return region



# Copyright 2009-2016 Yelp and Contributors
# Coypright 2017 Yelp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utility functions for MRJob
"""

# don't add imports here that aren't part of the standard Python library,
# since MRJobs need to run in Amazon's generic EMR environment
import contextlib
import glob
import itertools
import logging
import os
import pipes
import random
import shlex
import shutil
import sys
import tarfile
from copy import deepcopy
from datetime import timedelta
from distutils.spawn import find_executable
from logging import getLogger
from optparse import OptionParser
from zipfile import ZIP_DEFLATED
from zipfile import ZIP_STORED
from zipfile import ZipFile
from zipfile import is_zipfile

from mrjob.cat import bunzip2_stream
from mrjob.cat import decompress
from mrjob.cat import gunzip_stream
from mrjob.py2 import PY2
from mrjob.py2 import string_types

# these used to live in util; hush, pyflakes
bunzip2_stream
gunzip_stream

log = getLogger(__name__)


class NullHandler(logging.Handler):
    def emit(self, record):
        pass


def bash_wrap(cmd_str):
    """Escape single quotes in a shell command string and wrap it with ``bash
    -c '<string>'``.

    This low-tech replacement works because we control the surrounding string
    and single quotes are the only character in a single-quote string that
    needs escaping.

    .. deprecated:: 0.5.8
    """
    log.warning('bash_wrap() is deprecated and will be removed in v0.6.0')
    return "bash -c '%s'" % cmd_str.replace("'", "'\\''")


def buffer_iterator_to_line_iterator(chunks):
    """Alias for :py:func:`to_lines()`. Will be removed in v0.6.0.

    .. deprecated:: 0.5.0
    """
    log.warning('buffer_iterator_to_line_iterator() has been renamed to'
                ' to_lines(). This alias will be removed in v0.6.0')
    return to_lines(chunks)


def cmd_line(args):
    """build a command line that works in a shell.
    """
    args = [str(x) for x in args]
    return ' '.join(pipes.quote(x) for x in args)


def expand_path(path):
    """Resolve ``~`` (home dir) and environment variables in *path*.

    If *path* is ``None``, return ``None``.
    """
    if path is None:
        return None
    else:
        return os.path.expanduser(os.path.expandvars(path))


def file_ext(filename):
    """return the file extension, including the ``.``

    >>> file_ext('foo.tar.gz')
    '.tar.gz'
    """
    dot_index = filename.find('.')
    if dot_index == -1:
        return ''
    return filename[dot_index:]


def log_to_null(name=None):
    """Set up a null handler for the given stream, to suppress
    "no handlers could be found" warnings."""
    logger = logging.getLogger(name)
    logger.addHandler(NullHandler())


def log_to_stream(name=None, stream=None, format=None, level=None,
                  debug=False):
    """Set up logging.

    :type name: str
    :param name: name of the logger, or ``None`` for the root logger
    :type stderr: file object
    :param stderr:  stream to log to (default is ``sys.stderr``)
    :type format: str
    :param format: log message format (default is '%(message)s')
    :param level: log level to use
    :type debug: bool
    :param debug: quick way of setting the log level: if true, use
                  ``logging.DEBUG``, otherwise use ``logging.INFO``
    """
    if level is None:
        level = logging.DEBUG if debug else logging.INFO

    if format is None:
        format = '%(message)s'

    if stream is None:
        stream = sys.stderr

    handler = logging.StreamHandler(stream)
    handler.setLevel(level)
    handler.setFormatter(logging.Formatter(format))

    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)


def parse_and_save_options(option_parser, args):
    """Return a map from option name (``dest``) to a list of the arguments
    in *args* that correspond to that *dest*.

    This won't modify *option_parser*.
    """
    arg_map = {}

    def sim_callback(option, opt_str, value, parser):
        dest = option.dest
        arg_map.setdefault(dest, [])

        arg_map[dest].append(opt_str)
        if isinstance(value, string_types):
            arg_map[dest].append(value)
        elif value:
            arg_map[dest].extend(value)

    sim_parser = OptionParser(add_help_option=False)

    # optparse is no longer being maintained, so it's safe to access
    # hidden methods and attributes
    for option in option_parser._get_all_options():
        sim_parser.add_option(
            *(option._short_opts + option._long_opts),
            dest=option.dest,
            nargs=option.nargs,
            action='callback',
            type=('string' if option.type else None),
            callback=sim_callback)

    sim_parser.parse_args(args)

    return arg_map


def random_identifier():
    """A random 16-digit hex string."""
    return '%016x' % random.randint(0, 2 ** 64 - 1)


def read_file(path, fileobj=None, yields_lines=True, cleanup=None):
    """Yields lines from a file, possibly decompressing it based on file
    extension.

    Currently we handle compressed files with the extensions ``.gz`` and
    ``.bz2``.

    :param string path: file path. Need not be a path on the local filesystem
                        (URIs are okay) as long as you specify *fileobj* too.
    :param fileobj: file object to read from. Need not be seekable. If this
                    is omitted, we ``open(path)``.
    :param yields_lines: Does iterating over *fileobj* yield lines (like
                         file objects are supposed to)? If not, set this to
                         ``False`` (useful for :py:class:`boto.s3.Key`)
    :param cleanup: Optional callback to call with no arguments when EOF is
                    reached or an exception is thrown.
    """
    # sometimes values declared in the ``try`` block aren't accessible from the
    # ``finally`` block. not sure why.
    f = None
    try:
        # open path if we need to
        if fileobj is None:
            f = open(path, 'rb')
        else:
            f = fileobj

        decompressed_f = decompress(f, path)

        if decompressed_f is f and yields_lines:
            # this could be important; iterating over to_lines(f) is about 8x
            # slower than iterating over f
            lines = f
        else:
            lines = to_lines(decompressed_f)

        for line in lines:
            yield line
    finally:
        try:
            if f and f is not fileobj:
                f.close()
        finally:
            if cleanup:
                cleanup()


def read_input(path, stdin=None):
    """Stream input the way Hadoop would.

    - Resolve globs (``foo_*.gz``).
    - Decompress ``.gz`` and ``.bz2`` files.
    - If path is ``'-'``, read from stdin
    - If path is a directory, recursively read its contents.

    You can redefine *stdin* for ease of testing. *stdin* can actually be
    any iterable that yields lines (e.g. a list).
    """
    if stdin is None:
        stdin = sys.stdin

    # handle '-' (special case)
    if path == '-':
        for line in stdin:
            yield line
        return

    # resolve globs
    paths = glob.glob(path)
    if not paths:
        raise IOError(2, 'No such file or directory: %r' % path)
    elif len(paths) > 1:
        for path in paths:
            for line in read_input(path, stdin=stdin):
                yield line
        return
    else:
        path = paths[0]

    # recurse through directories
    if os.path.isdir(path):
        for dirname, _, filenames in os.walk(path, followlinks=True):
            for filename in filenames:
                for line in read_input(os.path.join(dirname, filename),
                                       stdin=stdin):
                    yield line
        return

    # read from files
    for line in read_file(path):
        yield line


# Thanks to http://lybniz2.sourceforge.net/safeeval.html for
# explaining how to do this!
def safeeval(expr, globals=None, locals=None):
    """Like eval, but with nearly everything in the environment
    blanked out, so that it's difficult to cause mischief.

    *globals* and *locals* are optional dictionaries mapping names to
    values for those names (just like in :py:func:`eval`).
    """
    # blank out builtins, but keep None, True, and False
    safe_globals = {
        'False': False,
        'None': None,
        'True': True,
        '__builtin__': None,
        '__builtins__': None,
        'set': set
    }

    # xrange is range in Python 3
    if PY2:
        safe_globals['xrange'] = xrange
    else:
        safe_globals['range'] = range

    # PyPy needs special magic
    def open(*args, **kwargs):
        raise NameError("name 'open' is not defined")
    safe_globals['open'] = open

    # add the user-specified global variables
    if globals:
        safe_globals.update(globals)

    return eval(expr, safe_globals, locals)


@contextlib.contextmanager
def save_current_environment():
    """ Context manager that saves os.environ and loads
        it back again after execution
    """
    original_environ = os.environ.copy()

    try:
        yield

    finally:
        os.environ.clear()
        os.environ.update(original_environ)


@contextlib.contextmanager
def save_cwd():
    """Context manager that saves the current working directory,
    and chdir's back to it after execution."""
    original_cwd = os.getcwd()

    try:
        yield

    finally:
        os.chdir(original_cwd)


def shlex_split(s):
    """Wrapper around shlex.split(), but convert to str if Python version <
    2.7.3 when unicode support was added.
    """
    if sys.version_info < (2, 7, 3):
        return shlex.split(str(s))
    else:
        return shlex.split(s)


def strip_microseconds(delta):
    """Return the given :py:class:`datetime.timedelta`, without microseconds.

    Useful for printing :py:class:`datetime.timedelta` objects.
    """
    return timedelta(delta.days, delta.seconds)


def tar_and_gzip(dir, out_path, filter=None, prefix=''):
    """Tar and gzip the given *dir* to a tarball at *out_path*.

    If we encounter symlinks, include the actual file, not the symlink.

    :type dir: str
    :param dir: dir to tar up
    :type out_path: str
    :param out_path: where to write the tarball too
    :param filter: if defined, a function that takes paths (relative to *dir*
                   and returns ``True`` if we should keep them
    :type prefix: str
    :param prefix: subdirectory inside the tarball to put everything into (e.g.
                   ``'mrjob'``)
    """
    # might move this to tests/; we use it a few places there
    log.warning('tar_and_gzip() is deprecated and will be removed in v0.6.0')

    if not os.path.isdir(dir):
        raise IOError('Not a directory: %r' % (dir,))

    if not filter:
        filter = lambda path: True

    tar_gz = tarfile.open(out_path, mode='w:gz')

    for dirpath, dirnames, filenames in os.walk(dir, followlinks=True):
        for filename in filenames:
            path = os.path.join(dirpath, filename)
            # janky version of os.path.relpath() (Python 2.6):
            rel_path = path[len(os.path.join(dir, '')):]
            if filter(rel_path):
                # copy over real files, not symlinks
                real_path = os.path.realpath(path)
                path_in_tar_gz = os.path.join(prefix, rel_path)
                tar_gz.add(real_path, arcname=path_in_tar_gz, recursive=False)

    tar_gz.close()


def to_lines(chunks):
    """Take in data as a sequence of bytes, and yield it, one line at a time.

    Only breaks lines on ``\\n`` (not ``\\r``), and does not add
    a trailing newline.

    Optimizes for:

    * chunks bigger than lines (e.g. reading test files)
    * chunks that are lines (idempotency)
    """
    # list of chunks with no final newline
    leftovers = []

    for chunk in chunks:
        start = 0

        while start < len(chunk):
            end = chunk.find(b'\n', start) + 1

            if end == 0:  # no newlines found
                leftovers.append(chunk[start:])
                break

            if leftovers:
                leftovers.append(chunk[start:end])
                yield b''.join(leftovers)
                leftovers = []
            else:
                yield chunk[start:end]

            start = end

    if leftovers:
        yield b''.join(leftovers)


def unique(items):
    """Yield items from *item* in order, skipping duplicates."""
    seen = set()

    for item in items:
        if item in seen:
            continue
        else:
            yield item
            seen.add(item)


def unarchive(archive_path, dest):
    """Extract the contents of a tar or zip file at *archive_path* into the
    directory *dest*.

    :type archive_path: str
    :param archive_path: path to archive file
    :type dest: str
    :param dest: path to directory where archive will be extracted

    *dest* will be created if it doesn't already exist.

    tar files can be gzip compressed, bzip2 compressed, or uncompressed. Files
    within zip files can be deflated or stored.
    """
    if tarfile.is_tarfile(archive_path):
        with contextlib.closing(tarfile.open(archive_path, 'r')) as archive:
            archive.extractall(dest)
    elif is_zipfile(archive_path):
        with contextlib.closing(ZipFile(archive_path, 'r')) as archive:
            for name in archive.namelist():
                # the zip spec specifies that front slashes are always
                # used as directory separators
                dest_path = os.path.join(dest, *name.split('/'))

                # now, split out any dirname and filename and create
                # one and/or the other
                dirname, filename = os.path.split(dest_path)
                if dirname and not os.path.exists(dirname):
                    os.makedirs(dirname)
                if filename:
                    with open(dest_path, 'wb') as dest_file:
                        dest_file.write(archive.read(name))
    else:
        raise IOError('Unknown archive type: %s' % (archive_path,))


def which(cmd, path=None):
    """Like the UNIX which command: search in *path* for the executable named
    *cmd*. *path* defaults to :envvar:`PATH`. Returns ``None`` if no
    such executable found.

    This is basically ``shutil.which()`` (which was introduced in Python 3.3)
    without the *mode* argument. Best practice is to always specify *path*
    as a keyword argument.
    """
    if hasattr(shutil, 'which'):
        return shutil.which(cmd, path=path)
    elif path is None and os.environ.get('PATH') is None:
        # find_executable() errors if neither path nor $PATH is set
        return None
    else:
        return find_executable(cmd, path=path)


def zip_dir(dir, out_path, filter=None, prefix=''):
    """Compress the given *dir* into a zip file at *out_path*.

    If we encounter symlinks, include the actual file, not the symlink.

    :type dir: str
    :param dir: dir to tar up
    :type out_path: str
    :param out_path: where to write the tarball too
    :param filter: if defined, a function that takes paths (relative to *dir*
                   and returns ``True`` if we should keep them
    :type prefix: str
    :param prefix: subdirectory inside the tarball to put everything into (e.g.
                   ``'mrjob'``)
    """
    if not os.path.isdir(dir):
        raise IOError('Not a directory: %r' % (dir,))

    if not filter:
        filter = lambda path: True

    try:
        zip_file = ZipFile(out_path, mode='w', compression=ZIP_DEFLATED)
    except RuntimeError:  # zlib not available
        zip_file = ZipFile(out_path, mode='w', compression=ZIP_STORED)

    for dirpath, dirnames, filenames in os.walk(dir, followlinks=True):
        for filename in filenames:
            path = os.path.join(dirpath, filename)
            # janky version of os.path.relpath() (Python 2.6):
            rel_path = path[len(os.path.join(dir, '')):]
            if filter(rel_path):
                # copy over real files, not symlinks
                real_path = os.path.realpath(path)
                path_in_zip_file = os.path.join(prefix, rel_path)
                zip_file.write(real_path, arcname=path_in_zip_file)

    zip_file.close()


### deprecated code, to remove in v0.6.0 ###

def args_for_opt_dest_subset(option_parser, args, dests=None):
    """For the given :py:class:`OptionParser` and list of command line
    arguments *args*, yield values in *args* that correspond to option
    destinations in the set of strings *dests*. If *dests* is None, return
    *args* as parsed by :py:class:`OptionParser`.

    .. deprecated:: 0.5.8
    """
    log.warning('args_for_opt_dest_subset() is deprecated and will be'
                ' removed in v0.6.0')

    for dest, value in _args_for_opt_dest_subset(option_parser, args, dests):
        yield value


def _args_for_opt_dest_subset(option_parser, args, dests=None):
    """See docs for :py:func:`args_for_opt_dest_subset()`. This function allows
    us to write a compatibility wrapper for the old API
    (:py:func:`parse_and_save_options()`).
    """
    values = deepcopy(option_parser.get_default_values())
    rargs = [x for x in args]
    option_parser.rargs = rargs
    while rargs:
        arg = rargs[0]
        if arg == '--':
            del rargs[0]
            return
        elif arg[0:2] == '--':
            for item in _process_long_opt(option_parser, rargs, values, dests):
                yield item
        elif arg[:1] == '-' and len(arg) > 1:
            for item in _process_short_opts(option_parser, rargs, values,
                                            dests):
                yield item
        else:
            del rargs[0]


def _process_long_opt(option_parser, rargs, values, dests):
    """Mimic function of the same name in ``OptionParser``, capturing the
    arguments consumed in *arg_map*
    """
    arg = rargs.pop(0)

    # Value explicitly attached to arg?  Pretend it's the next
    # argument.
    if "=" in arg:
        (opt, next_arg) = arg.split("=", 1)
        rargs.insert(0, next_arg)
    else:
        opt = arg

    opt = option_parser._match_long_opt(opt)
    option = option_parser._long_opt[opt]

    # Store the 'before' value of *rargs*
    rargs_before_processing = [x for x in rargs]

    if option.takes_value():
        nargs = option.nargs
        if nargs == 1:
            value = rargs.pop(0)
        else:
            value = tuple(rargs[0:nargs])
            del rargs[0:nargs]
    else:
        value = None

    option.process(opt, value, values, option_parser)

    if dests is None or option.dest in dests:
        # Measure rargs before and after processing. Yield difference.
        length_difference = len(rargs_before_processing) - len(rargs)
        for item in [opt] + rargs_before_processing[:length_difference]:
            yield option.dest, item


def _process_short_opts(option_parser, rargs, values, dests):
    """Mimic function of the same name in ``OptionParser``, capturing the
    arguments consumed in *arg_map*
    """
    arg = rargs.pop(0)
    stop = False
    i = 1
    for ch in arg[1:]:
        opt = "-" + ch
        option = option_parser._short_opt.get(opt)
        i += 1                      # we have consumed a character

        # Store the 'before' value of *rargs*
        rargs_before_processing = [x for x in rargs]

        # We won't see a difference in rargs for things like '-pJSON', so
        # handle that edge case explicitly.
        args_from_smashed_short_opt = []

        if option.takes_value():
            # Any characters left in arg?  Pretend they're the
            # next arg, and stop consuming characters of arg.
            if i < len(arg):
                rargs.insert(0, arg[i:])
                args_from_smashed_short_opt.append(arg[i:])
                stop = True

            nargs = option.nargs
            if nargs == 1:
                value = rargs.pop(0)
            else:
                value = tuple(rargs[0:nargs])
                del rargs[0:nargs]

        else:                       # option doesn't take a value
            value = None

        option.process(opt, value, values, option_parser)

        if dests is None or option.dest in dests:
            # Measure rargs before and after processing. Yield difference.
            length_difference = len(rargs_before_processing) - len(rargs)
            for item in ([opt] + args_from_smashed_short_opt +
                         rargs_before_processing[:length_difference]):
                yield option.dest, item

        if stop:
            break


def populate_option_groups_with_options(assignments, indexed_options):
    """Given a dictionary mapping :py:class:`OptionGroup` and
    :py:class:`OptionParser` objects to a list of strings represention option
    dests, populate the objects with options from ``indexed_options``
    (generated by :py:func:`scrape_options_and_index_by_dest`) in alphabetical
    order by long option name. This function primarily exists to serve
    :py:func:`scrape_options_into_new_groups`.

    :type assignments: dict of the form ``{my_option_parser: ('verbose',
                       'help', ...), my_option_group: (...)}``
    :param assignments: specification of which parsers/groups should get which
                        options
    :type indexed_options: dict generated by
                           :py:func:`util.scrape_options_and_index_by_dest`
    :param indexed_options: options to use when populating the parsers/groups
    """
    log.warning('populate_option_groups_with_options() is deprecated and'
                ' will be removed in v0.6.0')

    for opt_group, opt_dest_list in assignments.items():
        new_options = []
        for option_dest in assignments[opt_group]:
            for option in indexed_options[option_dest]:
                new_options.append(option)
        # New options must be added using add_options() or they will not be
        # allowed by the parser on the command line
        opt_group.add_options(new_options)
        # Sort alphabetically for help
        opt_group.option_list = sorted(opt_group.option_list,
                                       key=lambda item: item.get_opt_string())


def scrape_options_and_index_by_dest(*parsers_and_groups):
    """Scrapes ``optparse`` options from :py:class:`OptionParser` and
    :py:class:`OptionGroup` objects and builds a dictionary of
    ``dest_var: [option1, option2, ...]``. This function primarily exists to
    serve :py:func:`scrape_options_into_new_groups`.

    An example return value: ``{'verbose': [<verbose_on_option>,
    <verbose_off_option>], 'files': [<file_append_option>]}``

    :type parsers_and_groups: :py:class:`OptionParser` or
                              :py:class:`OptionGroup`
    :param parsers_and_groups: Parsers and groups to scrape option objects from

    :return: dict of the form ``{dest_var: [option1, option2, ...], ...}``
    """
    log.warning('scrape_options_and_index_by_dest() is deprecated'
                ' and will be removed in v0.6.0')

    all_options = {}
    job_option_lists = [g.option_list for g in parsers_and_groups]
    for option in itertools.chain(*job_option_lists):
        other_options = all_options.get(option.dest, [])
        other_options.append(option)
        all_options[option.dest] = other_options
    return all_options


def scrape_options_into_new_groups(source_groups, assignments):
    """Puts options from the :py:class:`OptionParser` and
    :py:class:`OptionGroup` objects in `source_groups` into the keys of
    `assignments` according to the values of `assignments`. An example:

    :type source_groups: list of :py:class:`OptionParser` and
                         :py:class:`OptionGroup` objects
    :param source_groups: parsers/groups to scrape options from
    :type assignments: dict with keys that are :py:class:`OptionParser` and
                       :py:class:`OptionGroup` objects and values that are
                       lists of strings
    :param assignments: map empty parsers/groups to lists of destination names
                        that they should contain options for
    """
    log.warning('scrape_options_into_new_groups() is deprecated'
                ' and will be removed in v0.6.0')

    all_options = scrape_options_and_index_by_dest(*source_groups)
    return populate_option_groups_with_options(assignments, all_options)
